// written by Zheng Gengbin
// perform prefetch read in out of core in forall statement

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
extern "C" void exit(...);
#endif

#include <stdio.h>

#include "compilerApi.h"

#include "anls_att.h"

//extern "C" int strcasecmp(...);

#define PREFTECH_SDX  1050
#define PREFETCH_BASE 1060

#define MAX_PREFETCH  3

int check_prefetch(SgStatement *s)
{
  if (prefetchFlag == 0) return 0;
  SgStatement *body = s->childList1(0);
  Forall_header_def * fhdef = (Forall_header_def *) s->attributeValue(0, FHDER_Def);

  forall_reference *lhsfref = (forall_reference *) body->attributeValue(0, Forall_Ref);
  SgSymbol *lhsarraySym = lhsfref->getarraysymbol();

  int prefetchNum=0;
  for (int i=body->numberOfAttributes(Forall_Ref)-1; i>0; i--) 
  {
          // if there are so many array do prefetch, maybe efficiency is lost
	if (prefetchNum >= MAX_PREFETCH) break;
	int can_prefetch = TRUE;
        forall_reference *fref = (forall_reference *) body->attributeValue(i, Forall_Ref);
  	SgArrayRefExp * array = (SgArrayRefExp *)fref->getexpression();
	SgSymbol *arraysymb = fref->getarraysymbol();
	SgArrayType *arrayType = (SgArrayType *)arraysymb->type();

	if (!is_out_of_core(arraysymb)) continue;
	  // if lhs is considered as in_core
        if (array->numberOfAttributes(SET_INCORE) > 0) continue;

        int storage = getArrayStorage(arraysymb);

	  // check the fasted changed dimension is continous.
	int dim;
	if (storage == COLUMN_MAJOR) 
	   dim = 0;
	else
	   dim = arrayType->dimension()-1;

	{
	  int indexNumber = fref->indexNumberOnDim(dim);
	  if (indexNumber == -1) continue;
	    // check if the forall header have stride not equal to 1
	  SgExpression *strExp = fhdef->getindextriple(indexNumber, 2);
	  if (strExp)
	  {
	    if (strExp->isInteger()) {
		int str = strExp->valueInteger();
		if (str != 1) can_prefetch = FALSE;
	    }
	    else
		can_prefetch = FALSE;
	    }
	}
	
	  // I can only support this communicationtype now.
        switch (fref->getCommunication()) 
	{
	case NONE: 
	case SHIFT:
	case UNKNOW:
		break;
	case REMAP:
		if (!is_out_of_core(lhsarraySym))
		   can_prefetch = FALSE;	
		break;
	default:
		can_prefetch = FALSE;	
		break;
	}
	if (can_prefetch == TRUE)
	{
if (!silentFlag)
printf("DO prefetch:%s\n", arraysymb->identifier());
	     // add flag to indicate prefetch status.
           fref->getexpression()->addAttribute(DO_PREFETCH,NULL,0);
	   prefetchNum++;
	}
  }
  return 1;
}

int prefetch_ok(SgStatement *body, int counter)
{
	if (prefetchFlag == 0) return 0;

        forall_reference *fref = (forall_reference *) body->attributeValue(counter, Forall_Ref);
	SgExpression *exp = fref->getexpression();
	if (exp->numberOfAttributes(DO_PREFETCH) > 0) return 1;
	else return 0;
}

// obtain the index varibles of next loop
SgSymbol *get_next_loop_index(SgStatement *s, SgStatement *innerloop, SgStatement *firstloop, int looplevel)
{
     if (prefetchFlag == 0) return 0;

     ASSERT(isSgForStmt(innerloop));
     Forall_header_def * fhdef = (Forall_header_def *) s->attributeValue(0, FHDER_Def);
     int i;
     int stloop = fhdef->getindexnumber()-looplevel-1;
//printf("looplevel:%d, stloop:%d \n", looplevel, stloop);
     SgSymbol *oldsym = fhdef->getindexsymb(looplevel);
     SgSymbol *oldIndexSymb = s->getSymbolInCurrentScope(mixstring,"T",oldsym->identifier());
     SgExpression *oldIndex = new SgVarRefExp(oldIndexSymb);
     SgSymbol *newIndexSymb = s->getSymbolInCurrentScope(mixstring,"T",oldsym->identifier(),1);
     SgExpression *newIndex = new SgVarRefExp(newIndexSymb);

       // get information on do loop's lbound and ubound.
     SgExpression * doHeaderExp=innerloop->expr(0);
     SgExpression * lboundExp=doHeaderExp->lhs()->lhs()->copyPtr();
     SgExpression * uboundExp=doHeaderExp->lhs()->rhs()->copyPtr();

     SgStatement *indexStmt;
     SgSymbol *tiles = s->getSymbolInCurrentScope(mixstring,"ST",NULL, stloop);
     SgExpression *tileExp = new SgVarRefExp(tiles);
     SgExpression *incExp = &( *oldIndex + *tileExp);
       // increase pcrc_T_i by pcrc_ST?
     indexStmt = new SgAssignStmt(*newIndex, *incExp);
     
     if (looplevel > 0)  
     {
	int org_numberOfChildren=innerloop->numberOfChildrenList1();
	SgStatement * * children = new SgStatement * [org_numberOfChildren];

	for (i=0; i<org_numberOfChildren; i++)
        children[i]=innerloop->childList1(i);

	SgSymbol *innerIndex;
	for (i=0; i<org_numberOfChildren; i++)
	{
          if (children[i]->variant()==FOR_NODE)
          {
		// recursivley call
		// return inner loop index symbol
	      innerIndex = get_next_loop_index(s, children[i], firstloop, looplevel-1);
          }
	}
	delete [ ] children;
	  // only if inner loop index become 0, current loop can increase
	  // by loop step; otherwise use the value of pcrc_T_i 
        SgExpression *innerIndexExp = new SgVarRefExp(innerIndex);
        SgExpression *con = &( SgEqOp(*innerIndexExp, *new SgValueExp(0))); 
        SgStatement *sameStmt = new SgAssignStmt( *newIndex, *oldIndex);
        indexStmt = new SgIfStmt(*con, *indexStmt, *sameStmt);
     }
     firstloop->insertStmtBefore(*indexStmt, *firstloop->controlParent());

       // if larger than ubound, reset to lbound
     if (looplevel < fhdef->getindexnumber()-1)
     {
        SgExpression *rewindCon = &( *newIndex > *uboundExp );
        SgStatement *resetStmt = new SgAssignStmt( *newIndex, *lboundExp );
        indexStmt = new SgLogIfStmt(*rewindCon, *resetStmt);
        firstloop->insertStmtBefore(*indexStmt, *firstloop->controlParent());
     }

     return newIndexSymb;
}

// modify pcrc_sdxi 
int modify_sdx(SgStatement *s)
{
  if (prefetchFlag == 0) return 0;

  Forall_header_def * fhdef = (Forall_header_def *) s->attributeValue(0, FHDER_Def);
  ASSERT(fhdef);
  SgStatement *body = s->childList1(0);
  int i,j;
   
  SgSymbol *sdx0Symb = s->getSymbolInCurrentScope(mixstring,"sdx",NULL,0);
  SgStatement *sdx0Stmt = NULL;
  SgExpression *sdxRhs;
  for ( i=1; i<body->numberOfAttributes(Forall_Ref); i++) 
  {
	if (prefetch_ok(body, i) == 0) continue;

        forall_reference *fref = (forall_reference *)body->attributeValue(i, Forall_Ref);
   	if (fref->sameAs!=-1) continue;

	SgArrayRefExp *array = (SgArrayRefExp *)fref->getexpression();
	SgSymbol *arraySym = (SgSymbol *) fref->getarraysymbol();
	SgArrayType *arrayType = (SgArrayType *)arraySym->type();

	SgSymbol *sdxSymb = s->getSymbolInCurrentScope(mixstring,"sdx",NULL,i);
	SgStatement *indexStmt0 = NULL; 
	SgStatement *indexStmt;
	  // find the pcrc_sdx? needed
  	for (j=0;j<body->numberOfAttributes(INDEX_SDX); j++)
  	{
     	   SgStatement *indexstmt = (SgStatement *)body->attributeValue(j,INDEX_SDX);
	   SgStatement *stmt =  indexstmt;
     	   if (!isSgAssignStmt(stmt))
		stmt = stmt->childList1(0);
	   ASSERT(isSgAssignStmt(stmt));
     	     // get sdx?
     	   SgSymbol *lhsSymb = stmt->expr(0)->symbol();
	   if (strcasecmp(lhsSymb->identifier(), sdx0Symb->identifier()) == 0) {
	     	  // is pcrc_sdx0
		sdx0Stmt = indexstmt;
		sdxRhs = sdx0Stmt->expr(1);
	   }
	     // get one sdx?
	   if (strcasecmp(lhsSymb->identifier(), sdxSymb->identifier()) == 0) {
		indexStmt0 = indexstmt;
             	  // if we use sdx0, we have to copy and use own
                indexStmt = indexStmt0;
	          // replace sdx0 in statement
		if (sdx0Stmt)
	            indexStmt->replaceSymbByExp(*sdx0Symb, *sdxRhs->copyPtr());
     	        for (int loop=0; loop<fhdef->getindexnumber(); loop++) 
     	        {
	            SgSymbol *oldsym = fhdef->getindexsymb(loop);
//	            SgSymbol *newsym = s->getSymbolInCurrentScope(mixstring,"T",oldsym->identifier());
	            SgSymbol *nextsym = s->getSymbolInCurrentScope(mixstring,"T",oldsym->identifier(),1);
	            SgExpression *nextExp = new SgVarRefExp(nextsym);
//printf("replace %s by %s\n",oldsym->identifier(), nextsym->identifier());
	            indexStmt->replaceSymbByExp(*oldsym, *nextExp);
     	        }
		  // this statement is for use when insert prefetch read
	 	  // outside do loop.
		SgStatement *stmt_for_first = indexStmt->copyPtr();
		ASSERT(stmt_for_first);
		array->addAttribute(PREFTECH_SDX, stmt_for_first, sizeof(indexStmt));
	   }	// end of if
	}

	  // failed to find any pcrc_sdx?
	  // use pcrc_sdx0 expression and create own pcrc_sdx?.
	if (indexStmt0 == NULL) {
     	      	indexStmt0 = (SgStatement *)body->attributeValue(0,INDEX_SDX);
	        SgStatement *stmt = indexStmt0;
     	        if (!isSgAssignStmt(stmt))
		      stmt = stmt->childList1(0);
     	        SgSymbol *oldsdxSymb = stmt->expr(0)->symbol();
	        indexStmt = indexStmt0->copyPtr();
	        SgExpression *sdxExp = new SgVarRefExp(sdxSymb);
	        indexStmt->replaceSymbByExp(*oldsdxSymb, *sdxExp);
		  // replace i bt pcrc_T_i1
     	        for (int loop=0; loop<fhdef->getindexnumber(); loop++) 
     	        {
	            SgSymbol *oldsym = fhdef->getindexsymb(loop);
	            SgSymbol *nextsym = s->getSymbolInCurrentScope(mixstring,"T",oldsym->identifier(),1);
	            SgExpression *nextExp = new SgVarRefExp(nextsym);
	            indexStmt->replaceSymbByExp(*oldsym, *nextExp);
     	        }
  	        body->addAttribute(INDEX_SDX, indexStmt, sizeof(indexStmt));
		array->addAttribute(PREFTECH_SDX, indexStmt->copyPtr(), sizeof(indexStmt));
	}
  }

  return 1;
}

// pcrc_base0 = pcrc_base1
// pcrc_base1 = mod(pcrc_base1+pcrc_in_core_size(pcrc_tdd1),pcrc_in_core_size(pcrc_tdd1)*2)
// 
int modify_mem_base(SgStatement *s, SgStatement *firstloop, SgSymbol *dad, int counter)
{
	  // we only need to set pcrc_base0 & pcrc_base1 once in the loop.
     	if (s->numberOfAttributes(PREFETCH_BASE) > 0) return 0;

	  // save pcrc_base1 => pcrc_base0
	SgSymbol *baseSym0=s->getSymbolInCurrentScope(mixstring,"base",NULL,0);
	SgExpression *baseExp0 = new SgVarRefExp(baseSym0);
	SgSymbol *baseSym1=s->getSymbolInCurrentScope(mixstring,"base",NULL,1);
	SgExpression *baseExp1 = new SgVarRefExp(baseSym1);
	SgAssignStmt *baseAss = new SgAssignStmt(*baseExp0, *baseExp1);
        firstloop->insertStmtBefore(*baseAss, *firstloop->controlParent());
	  // reset pcrc_base1
	SgSymbol *modFunc=s->getSymbolInCurrentScope("mod");
        SgFunctionRefExp *modExp = new SgFunctionRefExp (*modFunc);
	SgExpression *in_core_size = make_in_core_size(s, dad);
	SgExpression *incBase1 = &(*baseExp1 + *in_core_size);
	modExp->addArg(*incBase1);
	SgExpression *incBase2 = &(*in_core_size * *new SgValueExp(2));
	modExp->addArg(*incBase2);
	baseAss = new SgAssignStmt(*baseExp1, *modExp);
        firstloop->insertStmtBefore(*baseAss, *firstloop->controlParent());
	s->addAttribute(PREFETCH_BASE,NULL,0);
	return 1;
}

int get_previous_number(SgStatement *s, SgExpression *numberExp)
{
   for (int i=0; i<s->numberOfAttributes(OC_READ_NUMBER); i++)
   {
	SgExpression *number = (SgExpression *)s->attributeValue(i,OC_READ_NUMBER);
	if ( *number == *numberExp ) return 1;
   }
   return 0;
}

 // modify pcrc_oc_number_a ...
int modify_number(SgStatement *s, SgExpression *numberExp, int counter, int index, int loop)
{
//printf("modify_number counter:%d , loop:%d\n",counter,index);
	  // pcrc_oc_number_b modified to next tile.
  	Forall_header_def * fhdef = (Forall_header_def *) s->attributeValue(0, FHDER_Def);
	ASSERT(fhdef);
	SgSymbol *oldsym = fhdef->getindexsymb(index);
	SgSymbol *newsym = s->getSymbolInCurrentScope(mixstring,"T",oldsym->identifier());
//printf("newsym :%s\n",newsym->identifier());
	SgSymbol *nextsym = s->getSymbolInCurrentScope(mixstring,"T",oldsym->identifier(),1);
	SgExpression *nextExp = new SgVarRefExp(nextsym);
	numberExp->replaceSymbolByExpression(*newsym, *nextExp);
	return 1;
}

int add_first_prefetch_read(SgStatement *s, SgStatement *outloopStmt)
{
  if (prefetchFlag == 0) return 0;
  int i,j,k;
  SgStatement * body = s->childList1(0);
  Forall_header_def * fhdef = (Forall_header_def *) s->attributeValue(0, FHDER_Def);

  forall_reference *lhsfref = (forall_reference *) body->attributeValue(0, Forall_Ref);
  SgSymbol * lhsarraySym=(SgSymbol *) lhsfref->getarraysymbol();

    // ADD pcrc_base1 = 0 only once
  SgSymbol *baseSymb = s->getSymbolInCurrentScope(mixstring, "base", NULL, 1);
  SgExpression *baseExp = new SgVarRefExp(baseSymb);
  SgAssignStmt *baseAss = new SgAssignStmt(*baseExp, *new SgValueExp(0));
  outloopStmt->insertStmtBefore(*baseAss, *outloopStmt->controlParent());

  int sameAsi;
  for (i=1; i<body->numberOfAttributes(Forall_Ref); i++) 
  {
	if (prefetch_ok(body, i) == 0) continue;

        forall_reference *fref = (forall_reference *)body->attributeValue(i, Forall_Ref);
   	if (fref->sameAs!=-1) continue;
        if (fref->sameAs!=-1) sameAsi=fref->sameAs;
        else  sameAsi=i;

	SgArrayRefExp *array = (SgArrayRefExp *)fref->getexpression();
	SgSymbol * arraySym=(SgSymbol *) fref->getarraysymbol();
	SgArrayType * arrayType=(SgArrayType *)arraySym->type();

	  // ADD pcrc_oc_number_a(1) = ...
        for (k=0; k< array->numberOfAttributes(PREFETCH_NUM); k++)
	{
	  SgStatement *stmt;
	    // numStmt is the copyptr
  	  SgStatement *numStmt = (SgStatement *)array->attributeValue(k, PREFETCH_NUM);
	  ASSERT(numStmt);
	  stmt = numStmt ;
          if (!isSgAssignStmt(numStmt))
		numStmt = numStmt->childList1(0);
	  SgExpression *zero = new SgValueExp(0);
     	  for (int loop=0; loop<fhdef->getindexnumber(); loop++) 
     	  {
	   SgSymbol *oldsym = fhdef->getindexsymb(loop);
	   SgSymbol *newsym = s->getSymbolInCurrentScope(mixstring,"T",oldsym->identifier(),1);
	   SgExpression *newExp = new SgVarRefExp(newsym);
	   numStmt->replaceSymbByExp(*newsym, *zero);
     	  }
          outloopStmt->insertStmtBefore(*stmt, *outloopStmt->controlParent());
	}

  	SgStatement *ocPReadStmt0 = (SgStatement *)array->attributeValue(1, DO_PREFETCH);
	ASSERT (ocPReadStmt0);
	SgStatement *ocPReadStmt = ocPReadStmt0->copyPtr();

	  // ADD pcrc_sdx? = ...
	  // first find the pcrc_sdx? used regardless remap, none or shift
        SgFunctionCallExp *preadCall = isSgFunctionCallExp(ocPReadStmt0->expr(1));
	ASSERT(preadCall);
	SgExpression *sdxExp0 = preadCall->arg(2);
	SgSymbol * sdxSymb = sdxExp0->symbol();

	SgStatement  *indexStmt;
	SgStatement  *indexStmt0 = NULL;
  	for (j=0;j<array->numberOfAttributes(PREFTECH_SDX); j++)
	{
     	   SgStatement *indexStmt = (SgStatement *)array->attributeValue(j,PREFTECH_SDX);
	   SgExpression *zero = new SgValueExp(0);
     	   for (int loop=0; loop<fhdef->getindexnumber(); loop++) 
     	   {
	   SgSymbol *oldsym = fhdef->getindexsymb(loop);
	   SgSymbol *newsym = s->getSymbolInCurrentScope(mixstring,"T",oldsym->identifier());
	   indexStmt->replaceSymbByExp(*newsym, *zero);
	   newsym = s->getSymbolInCurrentScope(mixstring,"T",oldsym->identifier(),1);
	   indexStmt->replaceSymbByExp(*newsym, *zero);
	   }
	     // should use own pcrc_sdxi
	   SgSymbol *newSdx = s->getSymbolInCurrentScope(mixstring,"sdx",NULL,i);
	   if (strcasecmp(newSdx->identifier(), sdxSymb->identifier()) != 0)
	   {
	   SgExpression *newSdxExp = new SgVarRefExp(newSdx);
	   indexStmt->replaceSymbByExp(*sdxSymb, *newSdxExp);
	   ocPReadStmt0->replaceSymbByExp(*sdxSymb, *newSdxExp);
	   ocPReadStmt->replaceSymbByExp(*sdxSymb, *newSdxExp);
	   //sdxExp0->replace(*newSdxExp);
	   }
           outloopStmt->insertStmtBefore(*indexStmt, *outloopStmt->controlParent());
	}
     
/*
	  // ADD pcrc_base11 = 0
	SgSymbol *baseSymb = s->getSymbolInCurrentScope(mixstring, "base", NULL, 1);
	SgExpression *baseExp = new SgVarRefExp(baseSymb);
	SgAssignStmt *baseAss = new SgAssignStmt(*baseExp, *new SgValueExp(0));
        outloopStmt->insertStmtBefore(*baseAss, *outloopStmt->controlParent());
*/

	  // ADD pcrc_oc_prefetch_read
        if (  (is_out_of_core(lhsarraySym) ==0)
                       &&(fref->getCommunication()==UNKNOW))
	{
                SgSymbol* commSym=s->getSymbolInCurrentScope(mixstring, "comm",NULL,sameAsi);
                SgExpression * commExp= new SgVarRefExp(* commSym);
                commExp =& (SgNeqOp(*commExp ,*new SgValueExp(2)));
                ocPReadStmt = new SgIfStmt(*commExp,* ocPReadStmt);
	}
        outloopStmt->insertStmtBefore(*ocPReadStmt, *outloopStmt->controlParent());
  }
  return 1;
}


// combine same if condition statement
void combine_same_if(SgStatement *s)
{
      int org_numberOfChildren=s->numberOfChildrenList1();
      if (org_numberOfChildren == 0) return;
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i; 

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren; i++)
	combine_same_if(children[i]);

      delete [] children;
      
      org_numberOfChildren=s->numberOfChildrenList1();
      children = new SgStatement * [org_numberOfChildren];

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      SgStatement *previous = NULL;
      SgExpression *con = NULL;
      for (i=0; i<org_numberOfChildren; i++)
      {
	SgStatement *stmt = children[i];
	if ((stmt->variant() == LOGIF_STMT || stmt->variant() == LOGIF_NODE) &&
	     stmt->numberOfChildrenList2() == 0 ) 
	{
	   int reset = 0;
	   if (previous == NULL) {
		reset = 1;
	   }
	   else {
		if ( *con == *stmt->expr(0) ) {
		  // combine
//printf("Combine IF statements.\n");
		   for (int j=0; j<stmt->numberOfChildrenList1()-1; j++)
		   {
		      SgStatement *copy = stmt->childList1(j)->copyOnePtr();
		      SgStatement *last = previous->lastNodeOfStmt();
		      last->insertStmtBefore(*copy, *previous);
		   }
		   stmt->deleteStmt();
		   previous->setVariant(LOGIF_NODE);
		}
		else {
		   reset = 1;
		}
	   }
	   if (reset == 1) {
	   	previous = stmt;
	   	con = stmt->expr(0);
	   }
	}
	else {
		previous = NULL;
		con = NULL;
	}
      }
      delete [] children;
}



